﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

namespace LINQ
{
    class Program
    {
        static void Main(string[] args)
        {
            /*
             * A nyelvbe ágyazott lekérdezések (Language Integrated Query) lényege, hogy objektumorientált környezetben valósíthassunk 
             * meg lekérdezéseket gyűjteményeken, illetve gyűjteményszerű adattárakon (pl. adatbázis, leíró fájlok), hasonlóan ahhoz,
             * ahogy a relációs adatbázisok lekérdezései (SQL) működnek.
             */

            List<Int32> numList = new List<Int32> { 34, 23, 19, 14, 5, 41 };
            var numQuery1 = from num in numList // honnan
                            where num < 20 // feltétel
                            select num; // mit

            // A lekérdezés eredménye egy speciális gyűjtemény (IEnumerable) lesz, ennek köszönhetően a tartalmát felsorolhatjuk.
            Console.Write("numQuery1 result:");
            foreach(var num in numQuery1)
                Console.Write(" " + num);
            Console.WriteLine();
            
            // Lehetőségünk van az eredményeket sorba rendezni (orderby), egyéb értékeket definiálni (let), stb.
            var numQuery2 = from num in numList // lekérdezése forrása
                            let max = numList.Max() / 2 // új változó létrehozása
                            where num < max // feltétel
                            orderby num // rendezés
                            select max - num; // lekérdezés eredménye

            Console.Write("numQuery2 result:");
            foreach (var num in numQuery2)
                Console.Write(" " + num);
            Console.WriteLine();


            /*
             * Valójában a háttérben a System.Linq névtérben definiált kiegészítő metódusok futnak le a gyűjteményre,
             * amelyek lamda-kifejezést fogadnak paraméterként, és azt alkalmazzák az elemekre.
             */
            numQuery1 = numList
                .Where(num => num < 20)
                .Select(num => num);

            numQuery2 = numList
                .Select(num => new { Num = num, Max = numList.Max() / 2 })
                .Where(num => num.Num < num.Max)
                .OrderBy(num => num.Num)
                .Select(num => num.Max - num.Num);

            /*
             * Számos további kiegészítő metódus van még, néhány példa:
             *    sorba rendezés (OrderBy, OrderByDescending), 
             *    többszörös sorba rendezés (ThenBy),
             *    fordított sorred (Reverse),
             *    valahány elem vétele (Take) vagy kihagyása (Skip
             *    csoportosítás (GroupBy), konkatenálás (Concat)
             *    halmazműveletek (Union, Intersect, Except)
             *    összekapcsolások (Join, GroupJoin)
             *    statisztikai függvények (Sum, Avg, Min, Count, stb.)
             *    stb.
             */


            /*
             * A LINQ használatának előnye az úgynevezett késleltetett végrehajtás (deferred execution),
             * amely lehetővé teszi, hogy a kifejezés nem a híváskor, hanem az eredmény bejárásakor (a bejáró léptetésekor)
             * hajtódjon végre.
             * 
             * A lekérdezés valójában nem egyszerű felsoroló típust, hanem lekérdezés eredményt (IQueryable) ad, amely lényegében 
             * a kifejezést tárolja, viszont szintén bejárható (megvalósítja az IEnumerable-t).
             * 
             * Így bármikor módosítjuk a forrás gyűjteményt, vagy valamilyen külső változót,
             * a feldolgozáskor pedig a módosított gyűjteményen fut le a kiértékelés.
            */

            numList.AddRange(new Int32[] { 9, 38, 17 });

            Console.Write("numQuery1 result:");
            foreach (var num in numQuery1)
                Console.Write(" " + num);
            Console.WriteLine();

            Console.Write("numQuery2 result:");
            foreach (var num in numQuery2)
                Console.Write(" " + num);
            Console.WriteLine();
            
            /*
             * A késleltetett végrehajtás persze bizonyos esetekben ront a teljesítményen
             * (mivel minden bejáráskor lefut a kiértékelés). Ha azonnali végrehajtást szeretnénk,
             * az eredményt konvertálnunk kell.
             */
            var numSelection = numQuery1.ToList(); // var == List<Int32>

            const String fileName = "temp.txt";
            List<String> fileContent = new List<String>
            {
                "Lorem ipsum dolor sit amet consectetur adipiscing elit.",
                "Pellentesque sed sapien est.",
                "Fusce iaculis, sapien eu congue placerat, nunc ligula.",
                "Morbi venenatis congue fringilla.",
                "Vestibulum fringilla suscipit augue sagittis consequat.",
                "Nullam lobortis metus vitae erat pellentesque at molestie leo vehicula."
            };
            
            File.WriteAllLines(fileName, fileContent);
            using (FileEnumerable file = new FileEnumerable(fileName))
            {

                var wordQuery = file
                    .SelectMany(line => line.Split(new char[] { ' ' }))
                    .Select(word => word.ToLower().TrimEnd(new char[] { ',', '.' }))
                    .Distinct()
                    .OrderBy(word => word.Length)
                    .ThenBy(word => word)
                    .Aggregate((word1, word2) => word1 + ", " + Environment.NewLine + word2) + ".";

                Console.WriteLine("wordQuery result:");
                Console.WriteLine(wordQuery);
            }
            
            Console.ReadKey();
        }
    }
}
